/**
 * Combinatorical Functions for
 * Efficient Greedy Lottery Wheel Generator
 *
 * Author: Tyler McHenry <tyler@nerdland.net>
 *
 * For a more thorough description of the design and implementation of this
 * generator than can be provided in the in-line comments (with pretty diagrams
 * and everything!) please see "The Lottery Problem", a series of articles on
 * Nerdland: http://nerdland.net/the-lottery-problem/
 *
 * Licensing:
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * These functions deal with combinatorical computations and generations used in
 * setting up the data structures used by the wheel generator; they do not take
 * part in the actual wheel generation step itself.
 *
 * Tickets and matches are always represented by 64-bit bitfields, with the nth
 * least-significant bit set if and only if ball n is part of the ticket or
 * match in question. 
 *
 * These functions make use of pre-computed bit count information found in
 * bitcounts16.h.
 *
 * See the comments in wheeling.cc for more information on the generator.
 *
 * Last Modified: 11 July 2009
 *
 **/

#include <vector>
#include <sys/types.h>
#include "combinatorics.h"
#include "bitcounts16.h"

using std::vector;

/**
 * Quickly determines how many bits are set to 1 in a number using precomputed
 * information in the bitcount16.h header
 * */
static inline size_t countBits64(uint64_t num)
{
  return (BIT_COUNT_16[num & 0x00FFFF] +
          BIT_COUNT_16[(num >> 16) & 0x00FFFF] +
          BIT_COUNT_16[(num >> 32) & 0x00FFFF] + 
          BIT_COUNT_16[(num >> 48) & 0x00FFFF]);
}

/**
 * Generates all 64-bit numbers that have particular bit counts in each of their
 * four 16-bit words. The range parameter (r) limits the set bits to the
 * low-order r bits of the 64-bit numbers.
 * */
static void generateAllCombos(size_t bc_high_high_16, size_t bc_high_low_16,
                              size_t bc_low_high_16,  size_t bc_low_low_16,
                              size_t range, vector<uint64_t>& outvector)
{
  for (size_t idx_high_high = 0; idx_high_high < BIT_COUNT_16_REVERSE_IDX[bc_high_high_16];
       ++idx_high_high)
  {
    uint64_t high_high_16 = BIT_COUNT_16_REVERSE[bc_high_high_16][idx_high_high];

    if (range < 64 && high_high_16 != 0 &&
        (range <= 48 || ((0x0FFFF >> (16 - range - 48)) & high_high_16) != high_high_16)) continue;
        
    for (size_t idx_high_low = 0; idx_high_low < BIT_COUNT_16_REVERSE_IDX[bc_high_low_16];
         ++idx_high_low)
    {
      uint64_t high_low_16 = BIT_COUNT_16_REVERSE[bc_high_low_16][idx_high_low];

      if (range < 48 && high_low_16 != 0 &&
          (range <= 32 || ((0x0FFFF >> (16 - range - 32)) & high_low_16) != high_low_16)) continue;

      for (size_t idx_low_high = 0; idx_low_high < BIT_COUNT_16_REVERSE_IDX[bc_low_high_16];
           ++idx_low_high)
      {
        uint64_t low_high_16 = BIT_COUNT_16_REVERSE[bc_low_high_16][idx_low_high];

        if (range < 32 && low_high_16 != 0 &&
            (range <= 16 || ((0x0FFFF >> (16 - range - 16)) & low_high_16) != low_high_16)) continue;
        
        for (size_t idx_low_low = 0; idx_low_low < BIT_COUNT_16_REVERSE_IDX[bc_low_low_16];
             ++idx_low_low)
        {
          uint64_t low_low_16 = BIT_COUNT_16_REVERSE[bc_low_low_16][idx_low_low];

          if (range < 16 && low_low_16 != 0 &&
              (range <= 0 || ((0x0FFFF >> (16 - range - 0)) & low_low_16) != low_low_16)) continue;
         
          outvector.push_back((high_high_16 << 48) | (high_low_16 << 32) | (low_high_16 << 16) | low_low_16);
        }
      }
    }
  }
}

/**
 * Computes the binomial coefficient n-choose-k
 * */
size_t comb(size_t n, size_t k)
{
  size_t b[n+1];

  b[0] = 1;

  for (size_t i = 1; i <= n; ++i) {
    b[i] = 1;
    for (size_t j = i - 1; j > 0; --j) {
      b[j] += b[j-1];
    }
  }

  return b[k];
}

/**
 * Generates all tickets (or matches, which are like partial tickets) that can
 * exist in a t-from-r lottery. 
 * */
void generateAllTickets(size_t ticket_size, size_t range,
                        vector<uint64_t>& outvector)
{
  size_t max_bc_low_low_16 = std::max(0, std::min<int>(range, 16));
  size_t max_bc_low_high_16 = std::max(0, std::min<int>(range - 16, 16));
  size_t max_bc_high_low_16 = std::max(0, std::min<int>(range - 32, 16));
  size_t max_bc_high_high_16 = std::max(0, std::min<int>(range - 48, 16));
  
  for (size_t bc_high_high_16 = 0; bc_high_high_16 <= max_bc_high_high_16; ++bc_high_high_16)
  {
    for (size_t bc_high_low_16 = 0; bc_high_low_16 <= max_bc_high_low_16; ++bc_high_low_16)
    {
      for (size_t bc_low_high_16 = 0; bc_low_high_16 <= max_bc_low_high_16; ++bc_low_high_16)
      {
        for (size_t bc_low_low_16 = 0; bc_low_low_16 <= max_bc_low_low_16; ++bc_low_low_16)
        {
          if (bc_high_high_16 + bc_high_low_16 + bc_low_high_16 + bc_low_low_16 == ticket_size) {
            generateAllCombos(bc_high_high_16, bc_high_low_16, bc_low_high_16, bc_low_low_16,
                              range, outvector);
          }
        }
      }
    }
  }
}

/** Retrieves the w-th least significant 16-bit word for a 64-bit value */
static inline uint16_t getWord(uint64_t x, size_t w)
{
  return static_cast<uint16_t>(x >> (w * 16));
}

/** 
 * Generates all combinations of combo_size 1-bits that are set in the base-value.
 * If base_value describes a ticket, then this will generate all (combo-size)-matches
 * that match the ticket.
 * */
void generateAllCombos(uint64_t base_value, size_t combo_size, vector<uint64_t>& outvector)
{
  size_t base_size = countBits64(base_value);
  vector<uint64_t> masks;
  
  masks.reserve(base_size);

  // Generates masks, a vector of individual bitmasks for each bit set in the base_value
  for (size_t word_idx = 0; word_idx < 4; ++word_idx) {
    uint16_t word = getWord(base_value, word_idx);
    for (size_t word_mask_idx = 0; word_mask_idx < BIT_MASKS_16_IDX[word]; ++word_mask_idx) 
      { 
        uint64_t word_mask = BIT_MASKS_16[word][word_mask_idx];
        masks.push_back(word_mask << (word_idx * 16));
      }
  }

  // Iterates through every combo_size combination of elements of the masks
  // vector and combines them into a new match for the outvector.
  // NOTE: Doesn't work for combo_size > 16, thus the restriction of match size < 16
  uint16_t t_bit_mask = (0x0FFFF >> (16 - base_size));
  for (size_t mask_selector_idx = 0; mask_selector_idx < BIT_COUNT_16_REVERSE_IDX[combo_size];
       ++mask_selector_idx)
    {
      uint16_t mask_selector = BIT_COUNT_16_REVERSE[combo_size][mask_selector_idx];
      if ((mask_selector & t_bit_mask) == mask_selector) {
        uint64_t match = 0;
        for (size_t bit_mask_idx = 0; bit_mask_idx < BITS_SET_16_IDX[mask_selector]; ++bit_mask_idx) {
          match |= masks[BITS_SET_16[mask_selector][bit_mask_idx]];
        }
        outvector.push_back(match);
      }
    }
}

/**
 * Computes the starting coverage potential of every ticket. This is equivalent
 * to the number of unique other tickets that share at least one appropriately
 * sized match with any given ticket. This number is the same for all tickets.
 *
 * This particular method of calculation is due to a 2008 paper by Jans &
 * Degraeve in the European Journal of Operational Research:
 * http://ideas.repec.org/a/eee/ejores/v186y2008i1p104-110.html
 **/
size_t computeStartingCoverage(size_t range, size_t ticket_size,
                               size_t match_size)
{ 
  size_t starting_coverage = 0;
  
  for (size_t k = match_size; k < ticket_size; ++k) {
    starting_coverage += comb(ticket_size, k) *
      comb(range - ticket_size, ticket_size - k);
  }

  return starting_coverage;
}
