/**
 * Tool to Generate Precomputed Bit Count Information for
 * Efficient Greedy Lottery Wheel Generator
 *
 * Author: Tyler McHenry <tyler@nerdland.net>
 *
 * For a more thorough description of the design and implementation of this
 * generator than can be provided in the in-line comments (with pretty diagrams
 * and everything!) please see "The Lottery Problem", a series of articles on
 * Nerdland: http://nerdland.net/the-lottery-problem/
 *
 * Licensing:
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * This is a tool used to generate the bitcounts16.h header (sans comments) used
 * by combinatorics.cc in the wheel generator program. It is not necessary to
 * run this program yourself (its output is already included), but this tool is
 * included for completeness. The lookup tables it generates are:
 *
 * BIT_COUNT_16 : Maps a number to a count of bits set to 1 in the number's
 *                binary representation.
 *
 * BIT_COUNT_16_REVERSE : Maps a count of bits to an array of numbers with that
 *                        many bits set to 1 in their binary representation.
 *
 * BIT_COUNT_16_REVERSE_IDX : Maps a count of bits to the number of elements in
 *                            the array at that location in the
 *                            BIT_COUNT_16_REVERSE lookup table. 
 *
 * BIT_MASKS_16 : Maps a number to an array of bitmasks that each include a
 *                single bit that is set to 1 in the binary representation
 *                of the number.
 *
 * BIT_MASKS_16_IDX : Maps a number to the number of elementsin the array at that
 *                    location in the BIT_MASKS_16 lookup table.
 *
 * BITS_SET_16 : Maps a number to an array of the indices of the bits that are
 *               set to 1 in its binary representation, where the
 *               least-significant bit is index 0.
 *
 * BITS_SET_16_IDX : Maps a number to the number of elementsin the array at that
 *                   location in the BITS_SET_16 lookup table.
 *
 * Last Modified: 11 July 2009
 *
 **/

#include <iostream>
#include <cstring>

using namespace std;

static uint16_t BIT_COUNT_16_REVERSE[17][12870];
static size_t BIT_COUNT_16_REVERSE_IDX[17];

static uint16_t BIT_MASKS_16[65536][17];
static size_t BIT_MASKS_16_IDX[65536];

static uint16_t BITS_SET_16[65536][17];
static size_t BITS_SET_16_IDX[65536];

static uint8_t countBits16(uint16_t num)
{
  uint8_t count = 0;

  for (unsigned int i = 0; i < 16; ++i) {
    uint16_t mask = (1 << i);
    if ((num & mask) != 0) {
      ++count;
      BIT_MASKS_16[num][(BIT_MASKS_16_IDX[num])++] = mask;
      BITS_SET_16[num][(BITS_SET_16_IDX[num])++] = i;
    }
  }

  return count;
}

int main()
{
  memset(BIT_COUNT_16_REVERSE, '\0', sizeof(BIT_COUNT_16_REVERSE));
  memset(BIT_COUNT_16_REVERSE_IDX, '\0', sizeof(BIT_COUNT_16_REVERSE_IDX));

  memset(BIT_MASKS_16, '\0', sizeof(BIT_MASKS_16));
  memset(BIT_MASKS_16_IDX, '\0', sizeof(BIT_MASKS_16_IDX));

  memset(BITS_SET_16, '\0', sizeof(BITS_SET_16));
  memset(BITS_SET_16_IDX, '\0', sizeof(BITS_SET_16_IDX));

  cout << "#include <sys/types.h>" << endl;
  cout << endl;
  cout << "const uint8_t BIT_COUNT_16[] = {" << endl;
  cout << "  ";

  for (uint16_t i = 0; ; ++i) {
    uint8_t bit_count = countBits16(i); 
    cout << (uint16_t)bit_count;

    BIT_COUNT_16_REVERSE[bit_count][(BIT_COUNT_16_REVERSE_IDX[bit_count])++] = i;

    if (i != 65535) { 
      cout << ", ";
      if (i % 20 == 0) {
        cout << endl << "  ";
      }
    } else {
      break;
    }
  }

  cout << "};" << endl;

  cout << endl;
  cout << "const uint16_t BIT_COUNT_16_REVERSE[17][12870] = {" << endl;
  cout << "  ";
  for (size_t i = 0; i < 17; ++i) {
    cout << "  { ";
    for (size_t j = 0; j < BIT_COUNT_16_REVERSE_IDX[i]; ++j) {
      cout << BIT_COUNT_16_REVERSE[i][j] << ", ";
      if (j > 0 && j % 20 == 0) {
        cout << endl << "    ";
      }
    }
    cout << "}," << endl;
  }
  cout << "};" << endl;

  cout << endl;
  cout << "const uint16_t BIT_COUNT_16_REVERSE_IDX[17] = {";
  for (size_t i = 0; i < 17; ++i) {
    cout << BIT_COUNT_16_REVERSE_IDX[i] << ", ";
  }
  cout << "};" << endl;

  cout << endl;
  cout << "const size_t BIT_MASKS_16[65536][17] = {" << endl;
  cout << "  ";
  for (size_t i = 0; i < 65535; ++i) {
    cout << "  { ";
    for (size_t j = 0; j < BIT_MASKS_16_IDX[i]; ++j) {
      cout << BIT_MASKS_16[i][j] << ", ";
      if (j > 0 && j % 20 == 0) {
        cout << endl << "    ";
      }
    }
    cout << "}," << endl;
  }
  cout << "};" << endl;

  cout << endl;
  cout << "const size_t BIT_MASKS_16_IDX[65536] = {";
  for (size_t i = 0; i < 65535; ++i) {
    cout << BIT_MASKS_16_IDX[i] << ", ";
    if (i > 0 && i % 20 == 0) {
      cout << endl << "    ";
    }
  }
  cout << "};" << endl;

  cout << endl;
  cout << "const size_t BITS_SET_16[65536][17] = {" << endl;
  cout << "  ";
  for (size_t i = 0; i < 65535; ++i) {
    cout << "  { ";
    for (size_t j = 0; j < BITS_SET_16_IDX[i]; ++j) {
      cout << BITS_SET_16[i][j] << ", ";
      if (j > 0 && j % 20 == 0) {
        cout << endl << "    ";
      }
    }
    cout << "}," << endl;
  }
  cout << "};" << endl;

  cout << endl;
  cout << "const size_t BITS_SET_16_IDX[65536] = {";
  for (size_t i = 0; i < 65535; ++i) {
    cout << BITS_SET_16_IDX[i] << ", ";
    if (i > 0 && i % 20 == 0) {
      cout << endl << "    ";
    }
  }
  cout << "};" << endl;

  return 0;
}
