/**
 * Efficient Greedy Lottery Wheel Generator
 *
 * Author: Tyler McHenry <tyler@nerdland.net>
 *
 * For a more thorough description of the design and implementation of this
 * generator than can be provided in the in-line comments (with pretty diagrams
 * and everything!) please see "The Lottery Problem", a series of articles on
 * Nerdland: http://nerdland.net/the-lottery-problem/
 *
 * Licensing:
 * 
 *   This program is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 3 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Some general notes:
 *
 * - This is NOT an optimal wheel generator. It generates pretty good wheels,
 *   and it does so relatively quickly (given the problem complexity) but it
 *   does NOT reliably produce optimal results. This program is part of a
 *   continuing effort to produce such an optimal generator, but it does not yet
 *   achieve this goal.
 *
 * - This code has been tested only with relatively recent versions of g++, and
 *   not with any other compiler, or on Windows. This should in theory compile
 *   on any system that implements the C++ standard and which provides the POSIX
 *   sys/types.h header. Memory usage may increase on non-gcc builds due to the
 *   unavailability of gcc's struct-packing extensions.
 *
 * - This code makes use of the unordered_map container from Technical Report 1
 *   of the C++0x draft standard. If your compiler does not support this, you
 *   can safely replace all uses of std::tr1::unordered_map with std::map, at
 *   the expense of decreased performance (due to unnecessary balancing of the
 *   red-black tree that underlies std::map).
 *
 * - This code is optimized for performance, sometimes at the expense of
 *   readability. Since large data structures are dealt with when generating
 *   wheels for larger lotteries, many functions return their results through
 *   output parameters rather than as normal return values (to avoid copying).
 *
 * - The output of this program is a bit verbose, since for some inputs it may
 *   run for a very long time and it would be useful to know how far along it
 *   is. When building the problem graph, it outputs a . each time it finishes
 *   creating one thousand nodes. When generating the wheel, it outputs a . each
 *   time it covers 100 new tickets, and also outputs the percentage complete
 *   each time it increases by at least 1%. Percentage completeness is measured
 *   as the number of tickets covered out of all possible tickets. Note that the
 *   rate at which tickets are covered will increase as the computation
 *   progresses.
 *
 * Last Modified: 11 July 2009
 *
 **/

#include <cmath>
#include <cstdlib>
#include <iostream>
#include <vector>
#include <stack>
#include <string>
#include <sstream>
#include <fstream>
#include <memory>
#include <tr1/unordered_map>
#include <sys/types.h>
#include "wheelgen.h"
#include "combinatorics.h"

using std::cout;
using std::cerr;
using std::cin;
using std::endl;
using std::string;
using std::stringstream;
using std::ofstream;
using std::vector;
using std::stack;
using std::tr1::unordered_map;
using std::abs;
using std::auto_ptr;

// Problem parameters - used all over the place, so actually cleaner as
// unit-scope globals than as parameters to every function

static size_t g_range;        ///< Range size (r), e.g. a 49-ball lottery has a range of 49.
static size_t g_ticket_size;  ///< Ticket size (t), number of numbers on a ticket
static size_t g_match_size;   ///< Match size (m), number of numbers we want to gaurantee match the drawing)

/** Prints a ticket (or a match, which is like a partial ticket) in a nice, human readable format */
static string ticketToString(uint64_t ticket)
{
  stringstream sstr;
  for (size_t i = 0; i < 64; ++i) {
    if ((ticket & (1ull << i)) != 0) {
      if (sstr.str().length() > 0)  sstr << " ";
      sstr << i+1;
    }
  }

  return sstr.str();
}

/**
 * Builds the first part of the problem graph by creating a MatchNode struct for
 * each match in the all_matches vector.
 **/
static void convertMatchesToMatchNodes(vector<uint64_t>& all_matches,
                                       vector<MatchNode>& match_nodes,
                                       unordered_map<uint64_t, uint16_t>& match_idx_map)
{
  size_t ticket_idx_reserve_sz = comb(g_range - g_match_size, g_ticket_size - g_match_size);
  uint32_t match_idx = 0;

  match_nodes.resize(all_matches.size());

  for (vector<uint64_t>::const_iterator m = all_matches.begin(); 
       m != all_matches.end(); ++match_idx, ++m)
    {
      match_nodes[match_idx].ticket_indexes.reserve(ticket_idx_reserve_sz);
      match_idx_map[*m] = match_idx;

      if (match_idx % 1000 == 0) {
        cout << ".";
        cout.flush();
      }
    }
}


/**
 * Builds the second part of the problem graph by creating a TicketNode struct
 * for each ticket in the all_tickets vector. Then, it determines which matches
 * this ticket contains and "connects" the TicketNode to each of these match
 * nodes (by inserting an index reference to the ticket node in the match node
 * and vice-versa.
 **/
static void convertTicketsToTicketNodes(vector<uint64_t>& all_tickets,
                                        vector<TicketNode>& ticket_nodes,
                                        const unordered_map<uint64_t, uint16_t>& match_idx_map,
                                        vector<MatchNode>& match_nodes)
{
  size_t ticket_cover_initial_size = computeStartingCoverage(g_range, g_ticket_size, g_match_size);
  uint32_t ticket_idx = 0;

  ticket_nodes.resize(all_tickets.size());

  for (vector<uint64_t>::const_iterator t = all_tickets.begin(); 
       t != all_tickets.end(); ++ticket_idx, ++t)
    {
      vector<uint64_t> matches_for_ticket;
      generateAllCombos(*t, g_match_size, matches_for_ticket);
      
      ticket_nodes[ticket_idx].is_covered = 0;
      ticket_nodes[ticket_idx].remaining_coverage = ticket_cover_initial_size;
      ticket_nodes[ticket_idx].match_indexes.reserve(matches_for_ticket.size());

      for (vector<uint64_t>::const_iterator m = matches_for_ticket.begin();
           m != matches_for_ticket.end(); ++m)
        {
          uint16_t match_idx = (match_idx_map.find(*m))->second;
          ticket_nodes[ticket_idx].match_indexes.push_back(match_idx);
          match_nodes[match_idx].ticket_indexes.push_back(ticket_idx);
        }

      if (ticket_idx % 1000 == 0) {
        cout << ".";
        cout.flush();
      }

    }
}

/**
 * Builds the problem graph - converts ticket and match descriptions to
 * interconnected ticket and match nodes, and in the process loses the
 * description information.
 **/
static void generateNodes(vector<uint64_t>& all_tickets,
                          vector<uint64_t>& all_matches,
                          vector<TicketNode>& ticket_nodes,
                          vector<MatchNode>& match_nodes)
{
  unordered_map<uint64_t, uint16_t> match_idx_map(all_matches.size());

  cout << "Indexing all matches...";
  cout.flush();
  convertMatchesToMatchNodes(all_matches, match_nodes, match_idx_map);
  all_matches.clear();
  all_matches.reserve(0);
  cout << "done." << endl;


  cout << "Indexing all tickets...";
  cout.flush();
  convertTicketsToTicketNodes(all_tickets, ticket_nodes, match_idx_map, match_nodes);
  all_tickets.clear();
  all_tickets.reserve(0);
  cout << "done." << endl;
}


/**
 * This is the part that makes this generator fast. Here we do an efficient
 * depth-first walk of the problem graph and update the coverage potentials of
 * all tickets based on the ticket that we just selected for the wheel. Because
 * the coverage potentials are kept accurate, the selection of the next ticket
 * can always be done in O(n) time.
 **/
static void updatePotentials(vector<TicketNode>& ticket_nodes,
                             vector<MatchNode>& match_nodes,
                             uint32_t start_ticket_idx,
                             uint32_t& total_coverage,
                             size_t one_pct)
{
  stack<DFSStackEntry> dfs_stack;
  uint32_t current_level1_ticket_idx = 0;

  // Ideally this would be a bitset, but the size of a std::bitset must be
  // fixed at compile time, and if you're computing a very large wheel, the
  // size of this bitset can get huge. There's no reason to impose a huge
  // memory penalty on the generation of smaller wheels just for cleaner code,
  // and I didn't feel like pulling in boost just to get their dynamic
  // bitset. Remember that vector<bool> (as awful as it is design-wise) is
  // specialized to store internally as a bitset anyway.
  auto_ptr<vector<bool> > charged_level2_ticket_idxs(new vector<bool>(ticket_nodes.size()));

  dfs_stack.push(DFSStackEntry(start_ticket_idx, true, false));
  
  while (!dfs_stack.empty()) {
      
    DFSStackEntry& visit_node = dfs_stack.top();
    
    if (!visit_node.is_visited) {

      visit_node.is_visited = 1;

      if (visit_node.is_ticket) {
        
        TicketNode& t_node = ticket_nodes[visit_node.index];
        bool is_start_ticket = (visit_node.index == start_ticket_idx);

        if (!visit_node.is_level_two) {

          // First-level ticket node, i.e. a ticket that is covered by this
          // latest selection. It is in level 2 of the overall tree (if 0 is the
          // root).

          // Skip nodes that are already covered, because re-covering such
          // nodes should not reduce the coverage potential of anything else
          if (!is_start_ticket && t_node.is_covered) {
            dfs_stack.pop();
            continue;
          }

          // If the node is newly covered, put all of its children (which are
          // match nodes) into the DFS stack to be visited
          for (vector<uint16_t>::const_iterator m_idx = t_node.match_indexes.begin();
               m_idx != t_node.match_indexes.end(); ++m_idx)
            {
              dfs_stack.push(DFSStackEntry(*m_idx, false, !is_start_ticket));
            }

          if (!is_start_ticket) {

            // Mark node as covered
            t_node.is_covered = 1;
            
            // Charge for inability to cover self
            t_node.remaining_coverage -= std::min<uint32_t>(t_node.remaining_coverage, 1);

            // Produce progress output if it is time to do so
            if ((++total_coverage) % one_pct == 0) {
              cout << "\t " << (total_coverage * 100) / ticket_nodes.size() + 1 << "% done" << endl;
            }  else if (total_coverage % 100 == 0) {
              cout << ".";
              cout.flush();
            }

            // We are going to visit this node's subtree next, so set up the
            // context for doing that.
            current_level1_ticket_idx = visit_node.index;
            charged_level2_ticket_idxs.reset(new vector<bool>(ticket_nodes.size()));
            (*charged_level2_ticket_idxs)[visit_node.index] = true;
          }

        } else {
          
          // This would be a second-level ticket node, i.e. a ticket that is
          // having its coverage potential reduced as a result of the latest
          // selection. However the DFS traversal should never actually reach
          // here because we do the potential reductions from the match nodes,
          // to reduce the number of iterations through the DFS code. 

          cerr << "SHOULD NOT BE HERE!!" << endl;
        }
        
      } else {

        MatchNode& m_node = match_nodes[visit_node.index];
          
        if (!visit_node.is_level_two) {

          // First-level match node, i.e. a match contained by the just-selected
          // ticket.

          // Put each of its uncovered children (ticket nodes) into the DFS stack
          for (vector<uint32_t>::const_iterator t_idx = m_node.ticket_indexes.begin();
               t_idx != m_node.ticket_indexes.end(); ++t_idx)
            {
              TicketNode& t_node = ticket_nodes[*t_idx];

              // Charge for inability to cover root
              t_node.remaining_coverage -= std::min<uint32_t>(t_node.remaining_coverage, 1);

              // Only push the child if it is uncovered, since we need not do
              // any coverage reductions on the subtrees of already-covered
              // ticket nodes.
              if (!t_node.is_covered) {
                dfs_stack.push(DFSStackEntry(*t_idx, true, false));
              }
            }

        } else {

          // Second-level match node (appearing on level 3 of the overall tree),
          // i.e. a match contained by one of the tickets that was covered by
          // the most recent selection.

          // Charge its children. We can do this directly without needing to
          // push these children onto the stack. Each child represents a ticket
          // that is losing the potential to cover the newly-covered ticket that
          // is the root of the subtree we are currently traversing.
          for (vector<uint32_t>::const_iterator t_idx = m_node.ticket_indexes.begin();
               t_idx != m_node.ticket_indexes.end(); ++t_idx)
            {
              
              // Make sure we do not double-charge any one ticket within the
              // same subtree
              if (!(*charged_level2_ticket_idxs)[*t_idx]) {

                TicketNode& t_node = ticket_nodes[*t_idx];

                // Charge for inability to cover current level-2 (of the overall
                // tree) ticket
                t_node.remaining_coverage -= std::min<uint32_t>(t_node.remaining_coverage, 1);
                (*charged_level2_ticket_idxs)[*t_idx] = true;
              }
            }

          dfs_stack.pop();
        }
      }

    } else {

      // Have already traversed the subtree rooted at this node. Pop it and go
      // to its parent.
      
      dfs_stack.pop();
    }   
  }
}

/**
 * Generates a lottery wheel in a naiively greedy manner by continually
 * selecting the ticket with the largest remaining coverage potential, until the
 * largest remaining coverage potential is zero.
 **/
static void generateWheel(vector<uint64_t>& all_tickets,
                          vector<uint64_t>& all_matches,
                          vector<uint32_t>& wheel_ticket_idxs)
{
  vector<TicketNode> ticket_nodes;
  vector<MatchNode> match_nodes;

  uint32_t total_coverage = 0;
  bool is_first_iteration = true;
  bool done = false;
  size_t one_pct;

  generateNodes(all_tickets, all_matches, ticket_nodes, match_nodes);
  
  one_pct = std::max(ticket_nodes.size() / 100, static_cast<size_t>(1));

  cout << "Generating wheel..." << endl;

  while (!done) {

    uint32_t start_ticket_idx;

    if (is_first_iteration) {

      // In the first iteration, everything is symmetric, so we may as well just
      // select the ticket with index 0.
      
      start_ticket_idx = 0;
      is_first_iteration = false;

      cout << "\t Best coverage is " << ticket_nodes[0].remaining_coverage << " (by " 
           << ticket_nodes.size() << " tickets)" << endl;

    } else {

      start_ticket_idx = 0;

      uint32_t best_coverage_sz = 0;
      vector<uint32_t> best_coverage_ticket_idxs;

      // Do a linear scan of all remaining unselected tickets (including covered
      // tickets!) to determine which ticket (or set of tickets) have the
      // maximal coverage potential.
      
      for (uint32_t ticket_idx = 0; ticket_idx < ticket_nodes.size(); ++ticket_idx) {
        uint32_t remaining_coverage = ticket_nodes[ticket_idx].remaining_coverage;
        if (remaining_coverage > best_coverage_sz) {
          best_coverage_ticket_idxs.clear();
          best_coverage_ticket_idxs.push_back(ticket_idx);
          best_coverage_sz = remaining_coverage;
        } else if (remaining_coverage == best_coverage_sz) {
          best_coverage_ticket_idxs.push_back(ticket_idx);
        }
      }
      
      cout << "\t Next best coverage is " << best_coverage_sz << " (by " 
           << best_coverage_ticket_idxs.size() << " tickets)" << endl;

      if (best_coverage_sz == 0) {
        done = true;
        break;
      }

      // If more than one ticket has maximal coverage potential, select randomly
      // from among the set that do.

      start_ticket_idx = best_coverage_ticket_idxs[rand() % best_coverage_ticket_idxs.size()];
    }


    // Add the ticket to the wheel, zero out its coverage and count it as
    // covered.
    
    wheel_ticket_idxs.push_back(start_ticket_idx);
    ticket_nodes[start_ticket_idx].remaining_coverage = 0;
    ticket_nodes[start_ticket_idx].is_covered = 1;
    
    // Report progress if it is time to do so.

    if ((++total_coverage) % one_pct == 0) {
      cout << "\t " << (total_coverage * 100) / ticket_nodes.size() + 1 << "% done" << endl;
    } else if (total_coverage % 100 == 0) {
      cout << ".";
      cout.flush();
    }

    // Re-calculate the coverage potentials of tickets that are affected by this
    // selection.
    
    updatePotentials(ticket_nodes, match_nodes, start_ticket_idx,
                     total_coverage, one_pct);
    
    cout << "\t " << wheel_ticket_idxs.size() << " tickets selected for wheel." << endl;
  }

  cout << "Done." << endl;
}


int main()
{
  vector<uint64_t> all_tickets;
  vector<uint64_t> all_matches;

  // Prompt for problem parameters. In addition to sanity checks, we enforce the
  // following restrictions:
  //
  // 1. Range must be less than 64 (or else we cannot represent any given ticket
  //    in a uint64_t) 
  //
  // 2. The combination of range and ticket size cannot generate more than 2^30
  //    tickets (or else we cannot cram a DFStackEntry struct into 32 bits and
  //    our memory requirements explode)
  //
  // 3. Match sizes cannot be larger than 16, due to implementation details of
  //    generateAllCombos (in combinatorics.cc). Matches larger than 16 are
  //    pretty impractical anyway, so I saw no need to complicate that
  //    function's implementation to support such match sizes.
  
  cout << "Enter size of number range: ";
  cin >> g_range;

  if (g_range > 64) {
    cerr << "Ranges larger than 64 are not supported." << endl;
    return -1;
  }
  
  cout << "Enter size of tickets: ";
  cin >> g_ticket_size;

  if (g_ticket_size > g_range) {
    cerr << "Ticket size cannot be larger than range (that makes no sense!)" << endl;
    return -1;
  } else if ( comb(g_range, g_ticket_size) >= (1ul << 30) ) {
    cerr << "This combination of range and ticket size would generate too many tickets (" 
         << comb(g_range, g_ticket_size) << ")." << endl;
    cerr << "This program only supports quantities of tickets less than 2^30 (" 
         << (1ul << 30) << ")." << endl;
    return -1;
  }
  
  cout << "Enter size of matching to guarantee: "; 
  cin >> g_match_size;

  if (g_match_size > g_ticket_size) {
    cerr << "Match size cannot be larger than ticket size (that makes no sense!)" << endl;
    return -1;
  } else if (g_match_size > 16) {
    cerr << "Match sizes larger than 16 are not supported." << endl;
    return -1;
  }

  
  // First, generate all tickets and matches. When generateWheel is called, the
  // information about these tickets and matches will be forgotten to save
  // memory, but their relationship structure is preserved in the problem graph.
  
  cout << "Generating all possible tickets...";
  cout.flush();
  generateAllTickets(g_ticket_size, g_range, all_tickets);
  cout << "done. (generated " << all_tickets.size() << " tickets)" << endl;

  cout << "Generating all possible " << g_match_size << "-matches...";
  cout.flush();
  generateAllTickets(g_match_size, g_range, all_matches);
  cout << "done. (generated " << all_matches.size() << " " 
       << g_match_size << "-matches)" << endl;

  
  // Actually do the hard work of generating the wheel
  
  vector<uint32_t> wheel_ticket_idxs;
  generateWheel(all_tickets, all_matches, wheel_ticket_idxs); 


  // Re-generate all the ticket so we can convert the ticket indexes that
  // generateWheel gave us back into actual tickets with numbers on them. This
  // works since generateAllTickets always produces tickets in the same
  // deterministic order.
  
  cout << "Re-generating all possible tickets...";
  cout.flush();
  generateAllTickets(g_ticket_size, g_range, all_tickets);
  cout << "done. (generated " << all_tickets.size() << " tickets)" << endl;


  // Output the generated wheel to the screen and to a save file
  
  stringstream sstr_wheel_savename;

  sstr_wheel_savename << "wheel-" << g_range << "-" << g_ticket_size << "-" << g_match_size << ".txt";

  ofstream outfile(sstr_wheel_savename.str().c_str());      

  cout << "Wheel is: " << endl;
  for (vector<uint32_t>::const_iterator t_idx = wheel_ticket_idxs.begin();
       t_idx != wheel_ticket_idxs.end(); ++t_idx)
    {
      cout << "\t" << ticketToString(all_tickets[*t_idx]) << endl;
      outfile << ticketToString(all_tickets[*t_idx]) << endl;
    }
  
  cout << "Wrote wheel to file " << sstr_wheel_savename.str() << endl;      

  return 0;
}
